use std::path::Path;

use based::asset::AssetRoutes;
use based::page;
use based::request::{RequestContext, StringResponse};
use based::ui::components::{Breadcrumb, Shell, Tabs};
use based::ui::prelude::*;
use based::ui::primitives::flex::Row;
use based::ui::primitives::Optional;
use config::{get_prj, Project, ProjectConfig};
use maud::{PreEscaped, Render};
use rocket::fs::NamedFile;
use rocket::request::FromSegments;
use rocket::{get, launch, routes, State};
use site::gen_shell;

mod config;
mod invest;
mod site;

pub struct PathSegment {
    pub segments: Vec<String>,
}

impl PathSegment {
    pub fn to_str(&self) -> String {
        self.segments.join("/")
    }
}

impl<'r> FromSegments<'r> for PathSegment {
    type Error = ();

    fn from_segments(
        segments: rocket::http::uri::Segments<'r, rocket::http::uri::fmt::Path>,
    ) -> Result<Self, Self::Error> {
        let paths: Vec<_> = segments
            .filter_map(|x| {
                if x == "." {
                    return None;
                }

                if x == ".." {
                    return None;
                }

                Some(x.to_string())
            })
            .collect();

        Ok(PathSegment { segments: paths })
    }
}

#[get("/prj/<path..>")]
pub async fn project_page(
    path: PathSegment,
    ctx: RequestContext,
    shell: &State<Shell>,
    c: &State<ProjectConfig>,
) -> Option<StringResponse> {
    let mut prj = c.get(path.segments.first()?);
    for p in &path.segments[1..] {
        prj = prj.as_ref()?.sub.as_ref()?.get(p);
    }

    Some(render_project(shell, ctx, prj?, c, path.to_str()).await)
}

#[get("/static/<file>")]
async fn icon_res(file: &str) -> Option<NamedFile> {
    let path = Path::new("static").join(file);
    NamedFile::open(path).await.ok()
}

#[get("/")]
pub async fn main_page(
    c: &State<ProjectConfig>,
    ctx: RequestContext,
    shell: &State<Shell>,
) -> StringResponse {
    if c.len() == 1 {
        let keys = c.keys().collect::<Vec<_>>();
        let first = keys.first().unwrap();
        let prj = c.get(*first).unwrap();

        render_project(shell, ctx, prj, c, format!("{first}/")).await
    } else {
        // TODO : root project overview
        site::gen_site(c, ctx).await
    }
}

pub fn info_tab(prj: &Project, root: &str) -> PreEscaped<String> {
    Flex(
        Div()
            .vanish()
            .push(Rounded(
                Background(Padding(Code(&prj.description)).all(ScreenValue::_4)).color(Gray::_900),
            ))
            .push(Optional(prj.website.as_deref(), |website| {
                Link(website, Text(&format!("Website: {website}")))
            }))
            .push(Optional(prj.documentation.as_deref(), |docs| {
                Link(docs, Text(&format!("Documentation: {docs}")))
            }))
            .push(Optional(prj.since.as_deref(), |since| {
                Text(&format!("Since: {since}"))
            }))
            .push(Optional(prj.contact.as_ref(), |contact| {
                Div()
                    .vanish()
                    .push(Text("Contact").large())
                    .push(Text(&format!(
                        "eMail: {}",
                        contact
                            .email
                            .as_ref()
                            .map(|x| x.as_str())
                            .unwrap_or_default()
                    )))
            }))
            .push(Optional(prj.sub.as_ref(), |sub| {
                let mut prj_links = Vec::new();

                for key in sub.keys() {
                    prj_links.push(key);
                }

                Div()
                    .vanish()
                    .push(Text("Sub projects").large())
                    .push_for_each(&prj_links, |link: &_| {
                        Width(
                            ScreenValue::fit,
                            Button(Link(
                                &format!("/prj/{root}{link}"),
                                Text(&prj.sub.as_ref().unwrap().get(*link).unwrap().name),
                            )),
                        )
                    })
            })),
    )
    .gap(ScreenValue::_2)
    .direction(Direction::Column)
    .render()
}

fn up_to(keyword: &str, v: &[&str], separator: &str) -> String {
    let index = v.iter().position(|s| *s == keyword).unwrap_or(v.len());
    v[..index + 1].join(separator)
}

#[allow(non_snake_case)]
pub fn ProjectIcon(prj: &Project) -> PreEscaped<String> {
    Row(vec![
        Optional(prj.icon.as_ref(), |icon| Image(icon).alt("Project Icon")),
        Text(&prj.name)._2xl().render(),
    ])
    .gap(ScreenValue::_4)
    .full_center()
    .render()
}

pub async fn render_project(
    shell: &State<Shell>,
    ctx: RequestContext,
    prj: &Project,
    c: &State<ProjectConfig>,
    mut root: String,
) -> StringResponse {
    let title = format!("{} - Umbrella ☂️", prj.name);

    if !root.is_empty() && !root.ends_with('/') {
        root.push_str("/");
    }

    let root_paths: Vec<_> = root.split("/").filter(|x| !x.is_empty()).collect();

    let bc: Vec<_> = root_paths
        .iter()
        .map(|x| {
            let prj_path = up_to(x, &root_paths, "/");

            (
                ProjectIcon(&get_prj(c, &prj_path).unwrap()).0,
                format!("/prj/{}", prj_path),
            )
        })
        .collect();

    println!("{bc:?}");

    page!(
        shell,
        ctx,
        title,
        Div()
            .vanish()
            .push(Width(
                ScreenValue::fit,
                Margin(Breadcrumb(bc))
                    .top(ScreenValue::_6)
                    .bottom(ScreenValue::_2)
                    .x(ScreenValue::auto)
            ))
            .push(
                Padding(
                    Tabs()
                        .add_tab("info", Text("📜 Info").medium(), info_tab(prj, &root))
                        .add_tab(
                            "invest",
                            Text("💵 Invest").medium(),
                            Div().vanish().push(Text("TODO"))
                        )
                )
                .x(ScreenValue::_6)
            )
    )
}

// todo : fav icon

#[launch]
async fn rocket() -> _ {
    let conf_path: String = std::env::args()
        .skip(1)
        .next()
        .unwrap_or("./config.toml".to_string());
    let conf = config::Config::load(&conf_path);

    let shell = gen_shell();

    rocket::build()
        .mount_assets()
        .mount("/", routes![main_page, icon_res, project_page])
        .manage(conf)
        .manage(shell)
}
